%{
#include "../header.h"

bool WASERR = false;

int yyerror(char* s)
{
    err("%i: {%i} %s; lexema `%s'\n", yylineno, yynerrs, s, yytext);
}

set<char*,cmp<char*> > strings;
multimap<term,vector<term>,cmp_key<term> > predicates;
map<char*,int> vars;

const char* All = "_";
const char* Write = "write";

%}

%union{
struct term* Tree;
struct Vector* V;
}

%token ID IF VAR

%left ','

%type <Tree>    term id
%type <V>       list_term ruleset

%%
program:                list_predicat goal;

list_predicat:          predicat list_predicat | ;

term:                   id '(' list_term ')' { $1->v = $3->v;
                                                $1->var = $3->isvar ? -1:1;
                                                $$ = $1; delete $3;
                                        } |
                        id |
                        VAR {
                          $$ = creatterm(yytext);
                          $$->var = 0;
                          if( $$->name!=All )  vars[$$->name]++;
                        };

id:                     ID {
                          $$ = creatterm(yytext);
                          $$->var = 1;
                        };

list_term:              list_term ',' list_term
                        { $1->v.insert($1->v.end(),$3->v.begin(),$3->v.end());
                          $1->isvar = $1->isvar || $3->isvar;
                          $$=$1;
                          delete $3;
                        } |
                        term { $$ = new Vector; $$->v.push_back(*$1); $$->isvar = ($1->var==0);};

predicat:               term ruleset '.'
                        {
                            pair<term,vector<term> > p;
                            p.first = *$1;
                            p.second = $2->v;
                            predicates.insert(p);

                            for(map<char*,int>::iterator i = vars.begin(); i!=vars.end(); i++)
                             if( i->second == 1 ) err("%i: var `%s' uses 1th",yylineno,i->first);
                            vars.clear();
                        };

ruleset:                IF list_term { $$ = $2;} | {$$ = new Vector; };

goal:                   { err("And GOAL?!!");} |
                        ruleset  '.'
                        {
                         //for(map<char*,int>::iterator i = vars.begin(); i!=vars.end(); i++)
                         //if( i->second == 1 ) err("%i: var `%s' uses 1th",yylineno,i->first);
                         vars.clear();

                         if(WASERR) exit(1);

                         map<char*,term*> v;
                         for(vector<term>::iterator p = $1->v.begin(); p!=$1->v.end(); p++)
                         {
                          if( !call(*p,v) ){ printf("No.\n"); goto XXX;}
                         }

                         print_v(v);
                         printf("Yes.\n");

                         XXX: ;
                        };

%%

#define var(A)          (A.var==0)
#define novar(A)        (A.var)
#define atom(A)         (A.v.size()==0 && A.var)

term* creatterm(char* ss)
{
 term* x = new term;
 char* s = strdup(ss);
 pair<set<char*,cmp<char*> >::iterator,bool> r = strings.insert(s);
 if(!r.second) free(s);
 x->name = *r.first;
 return x;
}


/*
#define push(A,B)     {a=A; b=B; asm("pushl -4(%ebp)"); asm("pushl -8(%ebp)");}
#define pop(A,B)      {asm("popl -8(%ebp)"); asm("popl -4(%ebp)");}
int __noempty(){ asm("testl %ebp, %esp"); asm("jz ASL"); asm("movl $-1,%eax"); asm("jmp ASL2"); asm("ASL: xorl %eax, %eax"); asm("ASL2:");}
#define noempty       __noempty()
*/

#define push(A,B)    (iv[ic++]=A, iv[ic++]=B)
#define pop(A,B)    (B=iv[--ic], A=iv[--ic])
#define noempty     ic!=0

bool unify(term& A, term& B,map<char*,term*>& q1,map<char*,term*>& q2)
{
 int ic = 0;
 term* iv[2000];

 term* a, *b;

 push(&A,&B);

 while( noempty )
 {
  pop(a,b);

  if( a->var==0 && q1.count(a->name)==1 )  a = q1[a->name];
  if( b->var==0 && q2.count(b->name)==1 )  b = q2[b->name];

// out("a",*a); out("b",*b);
// getchar();

  if( a->var && b->var )
  {
   if(a->name != b->name ) return false;
   for(int i = 0; i<a->v.size(); i++){ push(&a->v[i],&b->v[i]);} //if( !unify(a->v[i],b->v[i],q1,q2) ) return false;
   continue;
  }

  if(a->name==All || b->name==All) continue;

  if( a->var==0 ) q1[a->name] = b;
  else
  if( b->var==0 ) q2[b->name] = a;
  else return false;
 }
 return true;
}

term* copy(term* x, map<char*,term*>& q)
{
 if( x->var==1 ) return x;
 if( x->var==0) return q[x->name];
 term* f = new term;
 for(int i=0; i<x->v.size(); i++)
 {
  term* n = copy(&x->v[i],q);
  if( n==NULL ){ delete f; return NULL;}
  f->v.push_back(*n);
 }
 f->name = x->name;
 f->var = 1;
 return f;
}

bool reset(map<char*,term*>& q1,map<char*,term*>& q2)
{
 term* x;

//printf("q1:"); print_v(q1);
//printf("q2:"); print_v(q2);
//printf("\n");

 for(map<char*,term*>::iterator k = q1.begin(); k!=q1.end(); k++)
 {
  if( k->second->var==0 && q2.count(k->second->name)==0 ){ /*printf("-f- q2\n");*/ return false;}
  if( k->second->var==0 )       x = q2[k->second->name];
  else                          x = k->second;
  k->second = copy(x,q2);
  if( k->second==NULL ){ /*printf("-f-NULL\n");*/ return false;}
 }
// printf("reset -T-\n");
 return true;
}

bool call(term& x,map<char*,term*>& v)
{
  map<char*,term*> v1, v2;

  if( x.name == Write ) if( x.v.size()!=0 ) return f_write(x,v);

  if( inbuild_is(x) ) return inbuild_call(x,v);

  multimap<term,vector<term>,cmp_key<term> >::iterator i;
  i = predicates.find(x);
  if( i == predicates.end() ){ err("predicat %s/%i not found",x.name,x.v.size()); return false;}

  for( ; i->first.name==x.name && i->first.v.size()==x.v.size() && i!=predicates.end(); i++)
  {
   v1.clear();
   v2.clear();
   v1 = v;

   bool b = unify(x,(term&) i->first,v1,v2);

   if( !b ) continue;

   for(vector<term>::iterator j = i->second.begin(); j!=i->second.end(); j++)
   {
    bool b = call(*j, v2);
    if( !b ){ goto false_ret;}
   }

  if( !reset(v1,v2) ){ goto false_ret;}

  v = v1;

  return true;
 }

false_ret:

 return false;
}

void print_v(map<char*,term*>& v)
{
 for(map<char*,term*>::iterator i = v.begin(); i!=v.end(); i++)
 {
  printf("%s=",i->first);
  print(*i->second);
 }
}

void printterm(term& x)
{
 if( atom(x) || var(x) ) printf("%s",x.name);
 else
 {
  printf("%s(",x.name);
  for(int i = 0; i<x.v.size(); i++)
  {
   printterm(x.v[i]);
   if( i+1 == x.v.size() )      printf(")");
   else                         printf(",");
  }
 }
}

void print(term& x)
{
 printterm(x);
 printf("\n");
}

void printpredicat(term& x,vector<term>& v)
{
 printterm(x); printf("%s", v.size() ? ":-":".\n");
 for(int i=0; i<v.size(); i++)
 {
  printterm(v[i]);
  if(i+1<v.size())      printf(",");
  else                  printf(".\n");
 }
}

void out(char* s,term& x)
{
 printf("%s: ",s);
 print(x);
}

int main(int ac,char** av)
{
 if(ac<2){ fprintf(stderr,"using: %s file\n",av[0]); exit(0);}
 
 
 
 yyin = fopen(av[1],"r");
 if(!yyin) perror(av[0]), exit(1);

 strings.insert((char*) All); // _
 strings.insert((char*) Write); // write(...)
 initial_inbuild(); // initial inbuild predicates

 yyparse();

 return 0;
}
